package backend;

/**
 * <p>berschrift: Java-Compiler fr PL0</p>
 * <p>Beschreibung: </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Organisation: </p>
 * @author Andreas, Thomas und Kai
 * @version 1.0
 */
import java.util.*;

import java.io.*;

import intermediate.*;
import intermediate.operators.*;

import backend.operators.*;


public class Assembler {

  private Map labelTable;
  private LinkedList frameStack;

  public void buildLabelTable() {
    Iterator it = asmList.iterator();
    AsmStatement stat;
    AsmLabel label;
    String labelName;

    labelTable = new HashMap();

    while(it.hasNext()){
      stat = (AsmStatement)it.next();

      if (stat instanceof AsmLabel) {
         label = (AsmLabel)stat;
         labelName = label.getName();
         labelTable.put(labelName,label);
         System.out.println("Label >"+label.getName()+"< eingefgt!");
      }
    }
  }





  private List optimizedList;
  private List asmList;
  private int maxTemp;

  public Assembler(List list) {
    optimizedList = list;
    maxTemp = 0;
    frameStack = new LinkedList();
    System.out.println("\nAssembler alive\n\n");
  }

  public void generateAsmStates(){

    Iterator it = optimizedList.iterator();
    Iterator it2;
    TreeNode instruc;
    List l;

    System.out.println("\ngeneriere Assembler-Code\n\n");

    asmList = new LinkedList();
    maxTemp = 0;

    // den eigenen Binrcode einschleusen
    asmList.add(new AsmLabel( "Error" ));
    asmList.add(new AsmPreDef( AsmPreDef.PrintErrorCode ));
    asmList.add(new AsmLabel( "Exit" ));
    asmList.add(new AsmPreDef( AsmPreDef.ExitProgCode ));
    asmList.add(new AsmLabel( "Input" ));
    asmList.add(new AsmPreDef( AsmPreDef.InputCode ));
    asmList.add(new AsmLabel( "Output" ));
    asmList.add(new AsmPreDef( AsmPreDef.OutputCode ));


    // durch die Codeblock-Liste iterieren
    while(it.hasNext()){
      l = (List)it.next();  // hiermit iteriere ich durch CODEBLCKE !!!

      // im Codeblock iterieren
      it2 = l.iterator();
      while (it2.hasNext()) {
        instruc = (TreeNode)it2.next();
        consumeImmStatement( instruc );
      }
    }

    System.out.print  ("\nAnzahl temporrer Variablen: ");
    System.out.println(maxTemp);
    System.out.println("\nAssembler-Code hat fertig\n\n");

  }

  public void generateAddresses(BinarFile binFile){
    Iterator it = asmList.iterator();
    AsmStatement stat;
    int codePos;
    int offset = binFile.getAssemblerOffset();

    System.out.println("\nadressiere Assembler-Code\n\n");


    // durch die Codeblock-Liste iterieren und Adressen vergeben
    codePos = offset;
    while(it.hasNext()){
      stat = (AsmStatement)it.next();
      stat.setPos(codePos);
      codePos += stat.getLength();
    }

    binFile.setAssemblerSize(codePos-offset);


    // Gre der Datensektion bestimmen
    binFile.setUserDataSize(4*maxTemp);

    System.out.print  ("\nGre Assembler-Code: ");
    System.out.println(codePos-offset);


    // Labels auflsen und Offset der User-Daten-Sektion bergeben
    buildLabelTable();

    AsmLabel label;
    AsmBranch branch;
    AsmCall call;
    AsmJump jump;
    AsmOutput output;
    int userDataOffset = binFile.getUserDataOffset();
    it = asmList.iterator();
    while(it.hasNext()){
      stat = (AsmStatement)it.next();
      stat.setUserDataOffset(userDataOffset);
      if (stat instanceof AsmBranch) {
         branch = (AsmBranch)stat;

         if ( labelTable.containsKey(branch.getTrueLabelName()) ) {
            label = (AsmLabel) labelTable.get(branch.getTrueLabelName());
            if (label!=null) {
               branch.setTrueLabel(label);
            }
            else System.out.println("LABELTABLE RETURNS NULL!");
         }
         else System.out.println("BRANCH-TARGET NOT IN LABELTABLE!");

         if ( labelTable.containsKey(branch.getFalseLabelName()) ) {
            label = (AsmLabel) labelTable.get(branch.getFalseLabelName());
            if (label!=null) {
               branch.setFalseLabel(label);
            }
            else System.out.println("LABELTABLE RETURNS NULL!");
         }
         else System.out.println("BRANCH-TARGET NOT IN LABELTABLE!");

/*         label = getAsmLabel(branch.getTrueLabelName());
         branch.setTrueLabel(label);
         label = getAsmLabel(branch.getFalseLabelName());
         branch.setFalseLabel(label);
         System.out.println("resolved: "+stat.toString());*/
      }
      if (stat instanceof AsmCall) {
         call = (AsmCall)stat;
         if ( labelTable.containsKey(call.getLabelName()) ) {
            label = (AsmLabel) labelTable.get(call.getLabelName());
            if (label!=null) {
               call.setTargetLabel(label);
            }
            else System.out.println("LABELTABLE RETURNS NULL!");
         }
         else System.out.println("CALL-TARGET NOT IN LABELTABLE!");
/*         label = getAsmLabel(call.getLabelName());
         call.setTargetLabel(label);
         System.out.println("resolved: "+stat.toString());*/
      }
      if (stat instanceof AsmJump) {
         jump = (AsmJump)stat;
         if ( labelTable.containsKey(jump.getLabelName()) ) {
            label = (AsmLabel) labelTable.get(jump.getLabelName());
            if (label!=null) {
               jump.setTargetLabel(label);
            }
            else System.out.println("LABELTABLE RETURNS NULL!");
         }
         else System.out.println("JUMP-TARGET NOT IN LABELTABLE!");
      }
      if (stat instanceof AsmOutput) {
         output = (AsmOutput)stat;
         if ( labelTable.containsKey(output.getLabelName()) ) {
            label = (AsmLabel) labelTable.get(output.getLabelName());
            if (label!=null) {
               output.setTargetLabel(label);
            }
            else System.out.println("LABELTABLE RETURNS NULL!");
         }
         else System.out.println("OUTPUT-TARGET NOT IN LABELTABLE!");
      }
    }


    if ( labelTable.containsKey("Entry") ) {
       label = (AsmLabel) labelTable.get("Entry");
       if (label!=null) {
          binFile.setEntryPoint(label.getPos());
          System.out.println("Label Entry an Code-Position "+label.getPos()+"!");
       } else {
         System.out.println("LABELTABLE RETURNS NULL!");
       }
    } else {
       System.out.println("ERROR: Label Entry not found!");
    }

/*  negative Zahlen getestet

   int i=-1;
   long l = (long)i;

   System.out.print("INTEGER: "); System.out.println(i);
   System.out.print("LONG:    "); System.out.println(l);

   byte[] bar = new byte[4];

   BinarFile.WriteLongToByteList(258,bar,0);
   System.out.println("258 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));

   BinarFile.WriteLongToByteList(2,bar,0);
   System.out.println("2 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));

   BinarFile.WriteLongToByteList(0,bar,0);
   System.out.println("0 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));

   BinarFile.WriteLongToByteList(-2,bar,0);
   System.out.println("-2 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));

   BinarFile.WriteLongToByteList(-258,bar,0);
   System.out.println("-258 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));

   BinarFile.WriteLongToByteList(-70000,bar,0);
   System.out.println("-70000 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));

   BinarFile.WriteLongToByteList(70000,bar,0);
   System.out.println("70000 = "+String.valueOf(bar[0])+"*"+String.valueOf(bar[1])+"*"+String.valueOf(bar[2])+"*"+String.valueOf(bar[3]));


   /*

   File f;
   FileOutputStream fos;
     try{
       f = new File("testfile.out");
       fos = new FileOutputStream(f);

       fos.write(bar);

       fos.close();
     }
     catch(Exception e)
     {
       System.out.print("CreateFile: ");
       System.out.println(e.getMessage());
     }


   //   System.out.print("INTEGER: "); System.out.println(i);
   //7System.out.print("LONG:    "); System.out.println(l);

  //*/


/*
    AsmLabel lab = this.getAsmLabel("Entry");
    if (lab==null) System.out.println("ERROR: Label Entry not found!");
    System.out.print("TEST: Label Entry an Code-Position "+lab.getPos()+"!");
*/
  }

  public List getAsmList() {
    return asmList;
  }

/*
  public AsmLabel getAsmLabel(String searched) {
    Iterator it = asmList.iterator();
    AsmStatement stat=null;
    boolean noAbort = true;
    boolean found = false;

    while(noAbort){
      stat = (AsmStatement)it.next();

      if (stat instanceof AsmLabel) {
         found = ((AsmLabel)stat).is(searched);
         if (found) System.out.print("FOUND   ");
         else System.out.print("NOT FOUND   ");
      }
      noAbort = it.hasNext();
      if (found) noAbort=false;
    }
    System.out.println("Exit loop");

    if (found) return ((AsmLabel)stat);
    else {
      System.out.println("Assembler:getAsmLabel(): Label >"+searched+"< not found!");
      return null;
    }
  }
*/

  private void registerTemp(int registee) {
    if (maxTemp < registee) maxTemp=registee;
  }

  private String getProcLabelName(int blocknumber) {
    return "Procedure " + String.valueOf(blocknumber);
  }

  private void consumeImmStatement(TreeNode statement) {
    //Expression e;
    //int value;

    if (statement instanceof MOVE) {
       consumeImmMOVE((MOVE)statement);
    } else

    if (statement instanceof FrameBuild) {
      consumeImmFrameBuild((FrameBuild)statement);
    } else

    if (statement instanceof FrameLeave) {
      consumeImmFrameLeave((FrameLeave)statement);
    } else

    if (statement instanceof CALL) {
      consumeImmCall((CALL)statement);
    } else

    if (statement instanceof LABEL) {
      asmList.add(new AsmLabel( ((LABEL)statement).getLabelName() ));
    } else

    if (statement instanceof INPUT) {
      consumeAsmInput( (INPUT)statement, Register.EAX );
    } else

    if (statement instanceof OUTPUT) {
      consumeAsmOutput( (OUTPUT)statement, Register.EAX );
    } else

    if (statement instanceof JUMP) {
      asmList.add(new AsmJump( ((JUMP)statement).getLabelName() ));
    } else

    if (statement instanceof CJUMP) {
      consumeImmCJUMP( (CJUMP)statement );
    } else

    {
      asmList.add(new InvalidImmCode("unknown statement"));
    }
  }

  private void consumeImmCONST (CONST constant, int targetReg) {
    asmList.add(new AsmLIT( targetReg, constant.getValue() ));
  }

  private void consumeImmFrameBuild (FrameBuild frame) {
    int frameNr = frame.getValue();
    frameStack.addLast(new Integer(frameNr));

    asmList.add( new AsmLabel( getProcLabelName(frameNr)) );
    if (frameNr==0) {
      asmList.add(new AsmLabel("Entry"));
      asmList.add(new AsmFBMain(frameNr));
    }
    else {
      asmList.add(new AsmFB(frameNr));
    }
  }

  private void consumeImmFrameLeave (FrameLeave frame) {
    int frameNr = frame.getValue();
    if (frameNr==0) {
      asmList.add(new AsmFLMain( frame.getValue() ));
      asmList.add(new AsmJump("Exit"));
    }
    else {
      asmList.add(new AsmFL( frame.getValue() ));
    }
    Integer frameValue = (Integer)frameStack.removeLast();
    if (frameValue.intValue() != frameNr) System.out.println("Assembler: FrameStackError");
  }

  private void consumeImmCall (CALL call) {
    Expression proc = call.getProc();
    int target, environement;

    if (proc instanceof CONST) {
      target = ((CONST)proc).getValue();
      environement = ((Integer)frameStack.getLast()).intValue();
      asmList.add(new AsmCall( getProcLabelName(target), (target==environement) ));
    }
    else {
      asmList.add(new InvalidImmCode("unknown construction: CALL("+proc.toString()+")"));
    }
  }

  private void consumeImmCJUMP (CJUMP cjump) {
    Expression e;

    // linken Operand laden
    e = cjump.getLeftExp();

    if (e instanceof CONST) {
      consumeImmCONST( (CONST)e, Register.EAX );
    } else

    if (e instanceof TEMP) {
      consumeLoadTEMP( (TEMP)e, Register.EAX );
    } else

    if (e instanceof MEM) {
      consumeLoadMEM( (MEM)e, Register.EAX );
    } else

    {
      asmList.add(new InvalidImmCode("CJUMP: invalid left operand"));
    }

    // Operand 2 laden
    e = cjump.getRightExp();

    if (e instanceof CONST) {
      consumeImmCONST( (CONST)e, Register.EBX );
    } else

    if (e instanceof TEMP) {
      consumeLoadTEMP( (TEMP)e, Register.EBX );
    } else

    if (e instanceof MEM) {
      consumeLoadMEM( (MEM)e, Register.EBX );
    } else

    {
      asmList.add(new InvalidImmCode("CJUMP: invalid right operand"));
    }

    // Vergleich ausfhren
    int relop = cjump.getRelopInt();
    switch (relop) {
        case CJUMP.EQ: asmList.add(new AsmCompare(Register.EAX, Register.EBX, Register.EAX, AsmCompare.EQ)); break;
        case CJUMP.NE: asmList.add(new AsmCompare(Register.EAX, Register.EBX, Register.EAX, AsmCompare.NE)); break;
        case CJUMP.LT: asmList.add(new AsmCompare(Register.EAX, Register.EBX, Register.EAX, AsmCompare.LT)); break;
        case CJUMP.GT: asmList.add(new AsmCompare(Register.EAX, Register.EBX, Register.EAX, AsmCompare.GT)); break;
        case CJUMP.LE: asmList.add(new AsmCompare(Register.EAX, Register.EBX, Register.EAX, AsmCompare.LE)); break;
        case CJUMP.GE: asmList.add(new AsmCompare(Register.EAX, Register.EBX, Register.EAX, AsmCompare.GE)); break;
    }

    // entsprechend verzweigen
    Label trueLabel = cjump.getTrueLabel();
    Label falseLabel = cjump.getFalseLabel();
    asmList.add(new AsmBranch(Register.EAX,trueLabel.toString(),falseLabel.toString()));

  }


  private void consumeImmBINOP (BINOP binop, int targetReg) {
    Expression e;

    // Operand 1 laden
    e = binop.getSrc1();

    if (e instanceof CONST) {
      consumeImmCONST( (CONST)e, Register.EAX );
    } else

    if (e instanceof TEMP) {
      consumeLoadTEMP( (TEMP)e, Register.EAX );
    } else

    if (e instanceof MEM) {
      consumeLoadMEM( (MEM)e, Register.EAX );
    } else

    {
      asmList.add(new InvalidImmCode("BINOP: invalid source 1"));
    }

    // Operand 2 laden
    e = binop.getSrc2();

    if (e instanceof CONST) {
      consumeImmCONST( (CONST)e, Register.EBX );
    } else

    if (e instanceof TEMP) {
      consumeLoadTEMP( (TEMP)e, Register.EBX );
    } else

    if (e instanceof MEM) {
      consumeLoadMEM( (MEM)e, Register.EBX );
    } else

    {
      asmList.add(new InvalidImmCode("BINOP: invalid source 2"));
    }

    // Operation ausfhren
    int op = binop.getOperatorInt();
    switch (op) {
        case 0: asmList.add(new AsmAdd(Register.EAX, Register.EBX)); break;
        case 1: asmList.add(new AsmSub(Register.EAX, Register.EBX)); break;
        case 2: asmList.add(new AsmMul(Register.EBX)); break;
        case 3: asmList.add(new AsmDiv(Register.EBX)); break;
        case 4: asmList.add(new NotImplemented("AND")); break;
        case 5: asmList.add(new NotImplemented("OR")); break;
        case 6: asmList.add(new NotImplemented("L-Shift")); break;
        case 7: asmList.add(new NotImplemented("R-Shift")); break;
        case 8: asmList.add(new NotImplemented("XOR")); break;
    }

    /*
     *  Das Ergebnis wird NICHT zurckgeschrieben, da BINOP keine eigen-
     *  stndige Anweisung mehr ist, sondern "nur" noch Teil der MOVE-
     *  Anweisung. Und diese bringt das speichern schon mit...
     *
     */
  }

  private void consumeAsmInput(INPUT input, int reg){
    asmList.add(new AsmInput(reg));
    consumeStoreMEM(reg, input.getAddress());
  }

  private void consumeAsmOutput(OUTPUT output, int reg){
    Expression e = output.getLoc();

    if (e instanceof TEMP) {
      consumeLoadTEMP( (TEMP)e, reg );
    } else
    if (e instanceof MEM) {
      consumeLoadMEM( (MEM)e, reg );
    } else
    if (e instanceof CONST) {
      consumeImmCONST( (CONST)e, reg );
    } else
    {
      asmList.add(new InvalidImmCode("OUTPUT: invalid expression"));
    }

    asmList.add(new AsmOutput(Register.EAX));
  }

  private void consumeLoadTEMP(TEMP temp, int targetReg){
    if (temp.getName().endsWith("fp")) {
      // Framepointer
      asmList.add(new AsmLoadFP(targetReg));
    } else {
      // normale Variable
      registerTemp(temp.getNumber());
      asmList.add(new AsmLoadTemp( temp.getNumber(), targetReg ));
    }
  }

  private void consumeLoadMEM(MEM mem, int targetReg){
//    asmList.add(new NotImplemented( "Load " + mem.toString() ));
    Expression exp = mem.getLoc();

    if (exp instanceof TEMP) {
      asmList.add(new AsmLoad( ((TEMP)exp).getNumber(), targetReg ));
    }
    else {
      asmList.add(new NotImplemented( "Load MEM(" + exp.toString() + ")" ));
    }
  }

  private void consumeStoreTEMP(int sourceReg, TEMP temp){
//    asmList.add(new NotImplemented("Store " + temp.toString()));
    if (temp.getName().endsWith("fp")) {
       asmList.add(new InvalidImmCode("Store sth to Temp fp"));
    } else {
      registerTemp(temp.getNumber());
      asmList.add(new AsmStoreTemp(sourceReg, temp.getNumber()));
    }
  }

  private void consumeStoreMEM(int sourceReg, MEM mem){
//    asmList.add(new NotImplemented("Store " + mem.toString()));
    Expression exp = mem.getLoc();

    if (exp instanceof TEMP) {
      asmList.add(new AsmStore( sourceReg, ((TEMP)exp).getNumber() ));
    }
    else {
      asmList.add(new NotImplemented( "Store MEM(" + exp.toString() + ")" ));
    }
  }




  private void consumeImmMOVE (MOVE instruction) {
    Expression e;
    int value;


    // von Quelle laden (in Register eax)
    e = (instruction).getSrc();

    if (e instanceof CONST) {
      consumeImmCONST( (CONST)e, Register.EAX );
    } else

    if (e instanceof BINOP) {
      consumeImmBINOP( (BINOP)e, Register.EAX );
    } else

    if (e instanceof TEMP) {
      consumeLoadTEMP( (TEMP)e, Register.EAX );
    } else

    if (e instanceof MEM) {
      consumeLoadMEM( (MEM)e, Register.EAX );
    } else

    {
      asmList.add(new InvalidImmCode("MOVE: invalid source"));
    }


    // an Ziel speichern (aus Register eax)
    e = ((MOVE)instruction).getDest();

    if (e instanceof TEMP) {
      consumeStoreTEMP( Register.EAX, (TEMP)e );
    } else

    if (e instanceof MEM) {
      consumeStoreMEM( Register.EAX, (MEM)e );
    } else

    {
      asmList.add(new InvalidImmCode("MOVE: invalid destination"));
    }

  }

}
