package backend.operators;

import backend.AsmStatement;
import backend.BinarFile;

public class AsmPreDef extends AsmStatement {

  public static final int PrintErrorCode = 0;
  public static final int ExitProgCode   = 1;
  public static final int InputCode      = 2;
  public static final int OutputCode     = 3;

  private int code;


  public AsmPreDef(int code) {
    super();
    this.code = code;
    switch (code) {
      case ExitProgCode: length=12; break;
      case OutputCode: length=79; break;
//      case OutputCode: length=1; break;
    }
  }

  public String toString() {
    String retval="PREDEFINED: invalid code";
    if (code==PrintErrorCode) retval="Print Error-Message";
    if (code==ExitProgCode) retval="Exit Programm";
    if (code==InputCode) retval="Read Integer";
    if (code==OutputCode) retval="Write Integer";

    return retval;
  }

  public byte[] getCode (byte[] list, int off) {
    if (code==ExitProgCode) {
       // Load 1 to eax
       list[off] = (byte)(0xb8);
       BinarFile.WriteLongToByteList(1, list, off+1);
       // Load 0 to ebx
       list[off+5] = (byte)(0xbb);
       BinarFile.WriteLongToByteList(0, list, off+6);
       // Int 0x80
       list[off+10] = (byte)(0xcd);
       list[off+11] = (byte)(0x80);
    } else

/* Test-Szenario
    if (code==OutputCode) {
      list[off] = (byte)(0xc3);  // ret
    }
//*/

    if (code==OutputCode) {
       /*
        *  Ausgangspunkt: Binrzahl in eax und alle Register gerettet...
        */

       // TEST (auskommentierter Code ist dann NOP...
       for (int i=0; i<this.length; i++) {
         list[off+i] = NOP;
       }


       /* alten esp merken */
       list[off  ] = (byte)(0x55);  // push ebp
       list[off+1] = (byte)(0x89);  // mov
       list[off+2] = (byte)(0xe5);  // esp=>ebp

       /* push 0x0 */
       list[off+3] = (byte)(0x6a);  // push (imm32)
       list[off+4] = (byte)(0x00);  // 0x0

       /* push \n */
       list[off+5] = (byte)(0x6a);  // push (imm8)
       list[off+6] = (byte)(0x0a);  // 0xa (\n)

       /* mov 0x0 => ebx */
       list[off+7] = (byte)(0xbb);  // mov ebx
       BinarFile.WriteLongToByteList(0, list, off+8);  // imm32

       /* if (eax<0) */
       list[off+12] = (byte)(0x83);  // cmp
       list[off+13] = (byte)(0xf8);  // eax
       list[off+14] = (byte)(0x00);  // 0x0
       list[off+15] = (byte)(0x7d);  // jge (jump if greater or equal)
       list[off+16] = (byte)(0x09);  // + 9 Bytes :AFTER THEN

       /* then */
       list[off+17] = (byte)(0x29);  // sub
       list[off+18] = (byte)(0xc3);  // eax from ebx
       list[off+19] = (byte)(0x89);  // mov
       list[off+20] = (byte)(0xd8);  // ebx to eax
       list[off+21] = (byte)(0xbb);  // mov ebx
       BinarFile.WriteLongToByteList(-1, list, off+22);  // imm32

       /* Label AFTER THEN */

       list[off+26] = (byte)(0xb9);  // mov ecx
       BinarFile.WriteLongToByteList(10, list, off+27);  // imm32

       /* Label WHILE */

       list[off+31] = (byte)(0xba);  // mov edx
       BinarFile.WriteLongToByteList(0, list, off+32);  // imm32
       list[off+36] = (byte)(0xf7);  // div
       list[off+37] = (byte)(0xf1);  // with divisor=ecx
       list[off+38] = (byte)(0x83);  // add
       list[off+39] = (byte)(0xc2);  // edx
       list[off+40] = (byte)(0x30);  // 0x30 (=48)
       list[off+41] = (byte)(0x52);  // push edx
       list[off+42] = (byte)(0x83);  // cmp
       list[off+43] = (byte)(0xf8);  // eax
       list[off+44] = (byte)(0x00);  // 0x0
       list[off+45] = (byte)(0x7f);  // jg (jump if greater)
       list[off+46] = (byte)(0xf0);  // - 16 Bytes :WHILE

       /* if (ebx==-1) */
       list[off+47] = (byte)(0x83);  // cmp
       list[off+48] = (byte)(0xfb);  // ebx
       list[off+49] = (byte)(0xff);  // -1
       list[off+50] = (byte)(0x75);  // jne (not equal)
       list[off+51] = (byte)(0x02);  // + 2 Bytes

       /* then push - */
       list[off+52] = (byte)(0x6a);  // push (imm8)
       list[off+53] = (byte)(0x3d);  // 0x2d (-)

       /* push <space> */
       list[off+54] = (byte)(0x6a);  // push (imm8)
       list[off+55] = (byte)(0x20);  // 0x20 (<space>)

       /* push ! */
       list[off+56] = (byte)(0x6a);  // push (imm8)
       list[off+57] = (byte)(0x21);  // 0x21 (!)

       /* printf */
       list[off+58] = (byte)(0xb8);  // mov eax
       BinarFile.WriteLongToByteList(4, list, off+59);  // imm32

       list[off+63] = (byte)(0xbb);  // mov ebx
       BinarFile.WriteLongToByteList(1, list, off+64);  // imm32

       list[off+68] = (byte)(0x89);  // mov
       list[off+69] = (byte)(0xe1);  // esp => ecx

       list[off+70] = (byte)(0x89);  // mov
       list[off+71] = (byte)(0xea);  // ebp => edx

       list[off+72] = (byte)(0x29);  // sub
       list[off+73] = (byte)(0xe2);  // esp from edx

       list[off+74] = (byte)(0xcd);  // int
       list[off+75] = (byte)(0x80);  // 0x80

       // man knnte Rckgabewert (eax) auslesen und ggf. Fehlermeldung ausgeben

       /* alten esp wieder aufsetzen */
       list[off+76] = (byte)(0x89);  // mov
       list[off+77] = (byte)(0xec);  // ebp=>esp
       list[off+78] = (byte)(0x5d);  // pop ebp

       /* zurckkehren */
       list[off+79] = (byte)(0xc3);  // ret

    } else

    {
      list = super.getCode(list, off);
    }
    return list;
  }

}
