package frontend.parsetree;

import antlr.Token;
import frontend.parser.PL0TokenTypes;
import intermediate.*;
import intermediate.operators.*;

public class BinrelNode extends PL0Node implements PL0TokenTypes {

    public BinrelNode(Token tok) {
        super(tok);
    }

    /**
     * Konvertiert den Knoten in ein Statement. Die Umwandlung erfolgt dabei so:
     *
     * Start:
     * a relop b (z.B. a < b)
     *
     * Ziel (am Beispiel fr relop = 'kleiner als' ):
     * CJUMP(<, a, b, t, f)
     *
     * Die Parameter a und b sind die zu vergleichenden Ausdrcke, t und f sind Label, die der
     * Methode bergeben werden.
     */
    public Statement convertConditional(Label t, Label f) {
        PL0Node left = (PL0Node)getFirstChild();
        PL0Node right = (PL0Node)left.getNextSibling();
        return new CJUMP(getNodeType(), left.convertExpression(), right.convertExpression(), t, f);
    };


    private int getNodeType() {
        int node = this.getType();
        switch (node) {
            case EQUALS: return CJUMP.EQ;
            case DIAMOND: return CJUMP.NE;
            case LT: return CJUMP.LT;
            case LEQ: return CJUMP.LE;
            case GT: return CJUMP.GT;
            case GEQ: return CJUMP.GE;
         }
        return -1;
    }

}
