package frontend.parsetree;

import antlr.Token;
import frontend.parser.PL0TokenTypes;
import intermediate.*;
import intermediate.operators.*;

public class OddNode extends PL0Node implements PL0TokenTypes {

    public OddNode(Token tok) {
        super(tok);
    }

    /**
     * Konvertiert den Knoten in ein Statement. Der Befehl ODD wird dabei umgesetzt als ein
     * binres AND mit dem Vergleichswert. Ist das Ergebnis 1, so war der Wert ungerade, ist
     * das Ergebnis 0, so war der Wert gerade.
     *
     * Start:
     * ODD a
     *
     * Ziel:
     * CJUMP(==, AND(a, 1), 1, t, f)
     *
     * Das Ziel ist ein geschachtelter Ausdruck mit folgender Bedeutung:
     * MOVE(TEMP, AND(a, 1));   //TEMP ist ein Zwischenspeicher
     * CJUMP(==, TEMP, 1, t, f)  //Wenn in TEMP 1 steht, war a ungerade
     */
    public Statement convertConditional(Label t, Label f) {
        PL0Node expr = (PL0Node)getFirstChild();
        Expression checkExpr = new BINOP(BINOP.AND, new CONST(1), expr.convertExpression());
        return new CJUMP(CJUMP.EQ, checkExpr, new CONST(1), t, f);
    };


    private int getNodeType() {
        int node = this.getType();
        switch (node) {
            case EQUALS: return CJUMP.EQ;
            case DIAMOND: return CJUMP.NE;
            case LT: return CJUMP.LT;
            case LEQ: return CJUMP.LE;
            case GT: return CJUMP.GT;
            case GEQ: return CJUMP.GE;
         }
        return -1;
    }

}
