package frontend.symbol;

import java.util.*;

/**
 * Eine Definition ist ein Eintrag in der Symboltabelle.Sie entspricht konzeptionell dem Symbol.
 * Das Symbol ist allerdings nur die Schnittstelle vom Token, die der Parser liefert. Es wird in eine
 * Definition umgewandelt. Damit findet eine Entkopplung von Parser und Symboltabelle statt.
 *
 * Fr jede Art von Bezeichner (Variable, Konstante, Prozedur) gibt es eine konkrete Unterklasse.
 */
public abstract class Definition {

    /**
     * Auftreten des Bezeichners im Quelltext
     */
    private Occurrence occurrence;

    /**
     * Sichtbarkeitsbereich des Bezeichners
     */
    private ScopeDef parentScope;

    /**
     * Name des Bezeichners
     */
    private String name;

    private int posInScope;

    /**
     * Konstruktor
     */
    public Definition(String name, Occurrence occ, ScopeDef parentScope) {
        this.occurrence  = occ;
        this.parentScope = parentScope;
        this.name        = name;
    }

    /**
     * Gibt den Typ des Bezeichners zurck. Da Definition eine abstrakte Oberklasse ist, wird diese
     * Methode hier nicht implementiert, das geschieht in konkreten Unterklassen.
     */
    public abstract int getType();

    /**
     * Gibt zurck, wo dieser Bezeichner im Quelltext gefunden wurde.
     */
    public Occurrence getOccurrence() {
        return occurrence;
    }

    /**
     * Gibt den Sichtbarkeitsbereich fr diesen Bezeichner zurck.
     */
    public ScopeDef getParentScope() {
        return parentScope;
    }

    /**
     * Setzt eine Referenz auf den Bezeichner, welcher dieses Symbol enthlt.
     */
    public void setParentScope(ScopeDef parentScope) {
        this.parentScope = parentScope;
    }

    /**
     * Liefert den Namen des Bezeichners zurck.
     */
    public String getName() {
        if (name == null) {
            return "~NO NAME~";
        } else {
            return name;
        }
    }

    public void setPosInScope(int pos) {
        posInScope = pos;
    }

    public int getPosInScope() {
        return posInScope;
    }


}
