package frontend.symbol.visualize;

import frontend.symbol.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.util.*;

public class JTableSymtabModel extends DefaultTableModel  {

    private SymbolTable symtab = SymbolTable.getSymbolTable();
    private ScopeDef currentScope;
    private List constants = new LinkedList();
    private List variables = new LinkedList();
    private List procedures = new LinkedList();
    private int nrVars;
    private int nrConst;
    private int nrProc;
    private Object[][] allDefs;

    /**
     * Konstruktor
     */
    public JTableSymtabModel(JTree tree) {
        //initiale Anzeige mit dem Basissichtbarkeitsbereich (Programmebene)
        currentScope = symtab.getBaseScope();
        //Wenn ein Knoten/Blatt im Baum selektiert wird, soll die Tabelle aktualisiert werden:
        tree.addTreeSelectionListener(new TreeSelectionListener() {
            public void valueChanged(TreeSelectionEvent e) {
                Object[] sdefs = e.getPath().getPath();
                //setze den Sichtbarkeitsbereich auf die ScopeDef, die angeklickt wurde
                currentScope = (ScopeDef)sdefs[sdefs.length - 1];
                //und aktualisiere die Matrix allDefs
                setupValues();
                //bringe die Tabelle dazu, sich neu zu zeichnen
                JTableSymtabModel.this.fireTableDataChanged();
            }
        });
        setupValues();
    }

    /**
     * Anzahl der Spalten
     */
    public int getColumnCount() {
        return 4;
    }

    /**
     * Anzahl der Zeilen entspricht Gre von allDefs-1. Das -1 deshalb, weil allDefs noch eine
     * Leerzeile am Ende hat, die nicht angezeigt werden soll.
     */
    public int getRowCount() {
        if (allDefs == null) {
            return 0;
        } else {
            return allDefs[0].length-1;
        }
    }

    /**
     * Spaltenberschriften
     */
    public String getColumnName(int i) {
        switch(i) {
            case 0: return "Konstanten";
            case 1: return "Variablen";
            case 2: return "Prozeduren";
            case 3: return "Definiert in";
        }
        return "";
    }

    /**
     * Methode des TableModels, zeigt die Matrix allDefs an, die von setupValues() aufgebaut wird
     */
    public Object getValueAt(int row, int column) {
        Definition d = (Definition)allDefs[column][row];
        if (d == null) return "";
        return d.getName();
    }


    /**
     * Erzeugt eine Matrix allDefs, die dann 1:1 von getValueAt() angezeigt wird. Dazu werden aus
     * dem aktuellen Sichtbarkeitsbereich alle Variablen, Konstanten und Prozeduren ermittelt und in
     * einem Array gespeichert. Dann wird zum nchsthheren Sichtbarkeitsbereich gegangen und das Gleiche
     * getan, solange bis alle Sichtbarkeitsbereiche abgearbeitet wurden. Die verschiedenen Arrays (fr
     * jeden Sichtbarkeitsbereich eines) werden dann zusammengesetzt.
     */
    private void setupValues() {
        int length = 0;
        //Liste aller Matrizen mit sichtbaren Konstanten, Variablen, Prozeduren
        List matrixes = new LinkedList();
        //Liste bereits gefundener Bezeichner, es soll nur das erste Auftreten eines Bezeichners
        //angezeigt werden, deshalb mssen wir uns merken, was wir schon gefunden haben
        List encountered = new LinkedList();
        do {
            Iterator it = currentScope.identifiers();
            variables.clear();
            constants.clear();
            procedures.clear();
            //Fr alle Identifier im Sichtbarkeitsbereich
            while (it.hasNext()) {
                Definition d = (Definition)it.next();
                //Wenn noch nicht gefunden, dann in der passenden Liste speichern
                if (!(encountered.contains(d.getName()))) {
                    encountered.add(d.getName());

                    if (d instanceof VariableDef) {
                        variables.add(d);
                    }
                    if (d instanceof ConstantDef) {
                        constants.add(d);
                    }
                    if (d instanceof ProcedureDef) {
                        procedures.add(d);
                    }
                }
            }
            nrVars = variables.size();
            nrConst = constants.size();
            nrProc = procedures.size();
            //Ntige Hhe der Matrix ermitteln, um alles anzeigen zu knnen
            int max = Math.max(Math.max(nrVars, nrConst), nrProc);
            //Falls etwas definiert wurde (meistens der Fall)
            if (max > 0) {
                //Erzeuge Matrix (mit Leerzeile am Ende) und flle die Listen ein
                Object[][] d = new Object[4][max + 1];
                for (int i = 0; i < max + 1; i++) {
                    if (i < constants.size()) {
                        d[0][i] = constants.get(i);
                    }
                    if (i < variables.size()) {
                        d[1][i] = variables.get(i);
                    }
                    if (i < procedures.size()) {
                        d[2][i] = procedures.get(i);
                    }
                }
                d[3][0] = currentScope;
                //zhle mit, wie gro die Gesamtmatrix sein muss
                length += max + 1;
                matrixes.add(d);
            }
            //weiter mit dem nchst hheren Sichtbarkeitsbereich
            currentScope = currentScope.getParentScope();
        //bis alle Sichtbarkeitsbereiche abgearbeitet sind
        } while (currentScope != null);


        Iterator it = matrixes.iterator();
        //Erzeuge die groe Matrix, die angezeigt werden soll
        allDefs = new Object[4][length];
        int dpos = 0;
        //fr alle Matrizen in der Liste: nimm eine Matrix und kopiere die Elemente nach allDefs
        while (it.hasNext()) {
            Object[][]o = (Object[][])it.next();
            for (int i = 0; i < o.length; i++) {
                for (int j = 0; j < o[0].length; j++) {
                    allDefs[i][j+dpos] = o[i][j];
                }
            }
            //merke, in welcher Zeile bei der nchsten Matrix fortzusetzen ist
            dpos += o[0].length;
        }

    }
}